﻿using Dapper;
using EVCB_OCPP.TaskScheduler.Models;
using Quartz;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading.Tasks;
using Microsoft.Data.SqlClient;
using Microsoft.Extensions.Logging;
using Microsoft.Extensions.Configuration;
using EVCB_OCPP.TaskScheduler.Helper;
using EVCB_OCPP.Domain;
using EVCB_OCPP.TaskScheduler.Services;
using Azure.Core;
using EVCB_OCPP.Domain.ConnectionFactory;
using System.Diagnostics;

namespace EVCB_OCPP.TaskScheduler.Jobs
{
    /// <summary>
    /// Update  EVSE's online status
    /// </summary>
    [DisallowConcurrentExecution]
    public class CheckEVSEOnlineJob : IJob
    {
        //private readonly string mainDBConnectString;
        //private readonly string onlineDBConnectString;

        public CheckEVSEOnlineJob(
            ILogger<CheckEVSEOnlineJob> logger,
            OnlineLogDbService onlineLogDbService,
            MainDbService mainDbService,
            ISqlConnectionFactory<MainDBContext> mainDbConnectionFactory,
            ISqlConnectionFactory<OnlineLogDBContext> onlineLogDbConnectionFactory)
        {
            this.logger = logger;
            this.onlineLogDbService = onlineLogDbService;
            this.mainDbService = mainDbService;
            this.mainDbConnectionFactory = mainDbConnectionFactory;
            this.onlineLogDbConnectionFactory = onlineLogDbConnectionFactory;
            //this.mainDBConnectString = configuration.GetConnectionString("MainDBContext");
            //this.onlineDBConnectString = configuration.GetConnectionString("OnlineLogDBContext");
        }

        private readonly ILogger logger;
        private readonly OnlineLogDbService onlineLogDbService;
        private readonly MainDbService mainDbService;
        private readonly ISqlConnectionFactory<MainDBContext> mainDbConnectionFactory;
        private readonly ISqlConnectionFactory<OnlineLogDBContext> onlineLogDbConnectionFactory;

        private DateTime latestHeartbeatTime = DateTime.Now;
        private List<EVSEOnlineRecord> updateData = new List<EVSEOnlineRecord>();
        private List<EVSEOnlineRecord> insertData = new List<EVSEOnlineRecord>();

        public async Task Execute(IJobExecutionContext context)
        {
            logger.LogDebug(this.ToString() + " :Starting........");
            try
            {
                List<EVSECurrentStatus> _EVSEs = await GetEVSEs();
                var checktime = DateTime.UtcNow.AddDays(-3);
                _EVSEs = _EVSEs.Where(x => x.HeartbeatUpdatedOn > checktime).ToList();
                List<Task> saveTasks = new List<Task>();

                var temp = _EVSEs.GroupBy(x => IsOnlineNow(x)).ToList();

                foreach (var g in temp)
                {
                    if (g.Key)
                    {
                        _EVSEs = g.Where(x => !x.Online).ToList();

                        for (int evseIndex = 0; evseIndex < _EVSEs.Count; evseIndex++)
                        {
                            var evse = _EVSEs[evseIndex];
                            //saveTasks.Add(UpdateEVSECurrentStatus(evse.CustomerId.ToString(), evse.ChargeBoxId, true, DefaultSetting.DefaultNullTime));
                            saveTasks.Add(mainDbService.UpdateEvseOnlineStatus(evse.ChargeBoxId, online: true));
                            //saveTasks.Add(UpdateOnlineRecords(evse.ChargeBoxId, true, evse.HeartbeatUpdatedOn, null));
                            saveTasks.Add(onlineLogDbService.InsertOnlineLog(evse.ChargeBoxId, evse.HeartbeatUpdatedOn));
                        }
                    }
                    else
                    {
                        _EVSEs = g.Where(x => x.Online).ToList();

                        for (int evseIndex = 0; evseIndex < _EVSEs.Count; evseIndex++)
                        {

                            var evse = _EVSEs[evseIndex];
                            //saveTasks.Add(UpdateEVSECurrentStatus(evse.CustomerId.ToString(), evse.ChargeBoxId, false, evse.HeartbeatUpdatedOn));
                            saveTasks.Add(mainDbService.UpdateEvseOnlineStatus(evse.ChargeBoxId, online: false, offlineTime: evse.HeartbeatUpdatedOn));
                            saveTasks.Add(onlineLogDbService.UpdateOnlineLog(evse.ChargeBoxId, evse.HeartbeatUpdatedOn));
                        }
                    }
                }

                await Task.WhenAll(saveTasks);
            }
            catch (Exception ex)
            {
                logger.LogDebug("ERROR " + this.ToString() + ex.ToString());
            }



            logger.LogDebug(this.ToString() + " :Finished........");

        }


        private async Task<List<EVSECurrentStatus>> GetEVSEs()
        {
            List<EVSECurrentStatus> result = new List<EVSECurrentStatus>();
            try
            {
                using var dbConn = await mainDbConnectionFactory.CreateAsync();

                string sqlstring =
                    """
                    SELECT m.CustomerId,m.Id,m.ChargeBoxId,m.Online,m.HeartbeatUpdatedOn,MachineConfigurations.ConfigureSetting HeartbeatInterval
                    FROM [dbo].[Machine] m left join [dbo].[MachineConfigurations] MachineConfigurations on m.ChargeBoxId = MachineConfigurations.ChargeBoxId
                    WHERE MachineConfigurations.ConfigureName = 'HeartbeatInterval'  and MachineConfigurations.ConfigureSetting!=''
                    """;
                result = (await dbConn.QueryAsync<EVSECurrentStatus>(sqlstring)).ToList();
            }
            catch (Exception ex)
            {
                logger.LogError("Query Data Error " + ex.ToString());
            }


            return result;
        }

        private async Task UpdateEVSECurrentStatus(string customerId, string ChargeBoxId, bool turnOn, DateTime offlineTime)
        {
            try
            {
                string sqlString = string.Format("UPDATE dbo.Machine SET Online=@Online {0} WHERE chargeBoxId=@chargeBoxId ", turnOn ? "" : " ,OfflineOn=@OfflineOn");
                using var dbConn = await mainDbConnectionFactory.CreateAsync();
                var parameters = new DynamicParameters();
                parameters.Add("@Online", turnOn, System.Data.DbType.Boolean);
                parameters.Add("@chargeBoxId", ChargeBoxId, System.Data.DbType.String, System.Data.ParameterDirection.Input, 50);

                if (!turnOn)
                {
                    parameters.Add("@OfflineOn", offlineTime, System.Data.DbType.DateTime);
                }

                await dbConn.ExecuteAsync(sqlString, parameters);
            }
            catch (Exception ex)
            {
                logger.LogError("Update Data Error " + ex.ToString());
            }
        }


        private async Task UpdateOnlineRecords(string chargeBoxId, bool online, DateTime hearbeatDt, Int64? rowId)
        {
            try
            {
                if (online)
                {
                    string sqlString = "INSERT INTO dbo.EVSEOnlineRecord (\"ChargeBoxId\",\"OnlineTime\",\"OfflineTime\")" +
                       "VALUES( @ChargeBoxId,@OnlineTime,@OfflineTime); ";
                    using var dbConn = await onlineLogDbConnectionFactory.CreateAsync();
                    {
                        var parameters = new DynamicParameters();
                        parameters.Add("@ChargeBoxId", chargeBoxId, System.Data.DbType.String, System.Data.ParameterDirection.Input, 36);
                        parameters.Add("@OnlineTime", hearbeatDt, System.Data.DbType.DateTime);
                        parameters.Add("@OfflineTime", DefaultSetting.DefaultNullTime, System.Data.DbType.DateTime);

                        await dbConn.ExecuteAsync(sqlString, parameters);
                    }
                }
                else
                {
                    if (rowId.HasValue)
                    {
                        string sqlString = "UPDATE dbo.EVSEOnlineRecord SET OfflineTime=@OfflineTime WHERE Id=@Id";
                        using (var dbConn = await onlineLogDbConnectionFactory.CreateAsync())
                        {
                            var parameters = new DynamicParameters();
                            parameters.Add("@OfflineTime", hearbeatDt, System.Data.DbType.DateTime);
                            parameters.Add("@Id", rowId.Value, System.Data.DbType.Int64);

                            await dbConn.ExecuteAsync(sqlString, parameters);
                        }
                    }
                    else
                    {
                        logger.LogError("Can't find update row id from " + chargeBoxId);
                    }

                }
            }
            catch (Exception ex)
            {
                logger.LogError(string.Format("UpdateOnlineRecords Exception:{0}", ex.ToString()));
            }




        }



        private async Task<EVSEOnlineRecord> GetOnlineRecord(string chargeBoxId)
        {
            EVSEOnlineRecord lastrow = new EVSEOnlineRecord();
            try
            {
                string sqlString = string.Format("SELECT Id FROM dbo.EVSEOnlineRecord WHERE chargeBoxId=@chargeBoxId Order by OnlineTime desc");
                using (var dbConn = await onlineLogDbConnectionFactory.CreateAsync())
                {
                    var parameters = new DynamicParameters();
                    parameters.Add("@chargeBoxId", chargeBoxId, System.Data.DbType.String, System.Data.ParameterDirection.Input, 36);


                    lastrow = await dbConn.QueryFirstOrDefaultAsync<EVSEOnlineRecord>(sqlString, parameters);
                    //lastrow = result.FirstOrDefault();
                }

            }
            catch (Exception ex)
            {
                logger.LogError("Query Data Error " + ex.ToString());
            }
            return lastrow;
        }

        private async Task<Dictionary<string, long?>> GetOnlineRecordId(List<string> chargeBoxIds)
        {
            Dictionary<string, long?> toReturn = new();
            Dictionary<string, long> queryResult;
            string sqlString = """
                WITH CTE AS (
                    SELECT
                        Id, ChargeBoxId,
                        ROW_NUMBER() OVER (PARTITION BY s.ChargeBoxId ORDER BY s.OnlineTime DESC) AS RowNum
                    FROM [dbo].[EVSEOnlineRecord] s
                    WHERE ChargeBoxId in @ChargeBoxIds
                )
                SELECT Id,ChargeBoxId
                FROM CTE
                WHERE RowNum = 1;
                """;
            var parameters = new DynamicParameters();
            parameters.Add("@ChargeBoxIds", chargeBoxIds.Select(x => new DbString() { Value = x, Length = 36 }));

            try
            {
                using (var dbConn = await onlineLogDbConnectionFactory.CreateAsync())
                {
                    var temp = await dbConn.QueryAsync<EVSEOnlineRecord>(sqlString, parameters);
                    queryResult = temp.ToDictionary(x => x.ChargeBoxId, x => x.Id);
                    //lastrow = result.FirstOrDefault();
                }

                toReturn = chargeBoxIds.ToDictionary<string, string, long?>(keySelector: x => x, elementSelector: x => queryResult.ContainsKey(x) ? queryResult[x] : null);
            }
            catch (Exception ex)
            {
                logger.LogError("Query Data Error " + ex.ToString());
                toReturn = chargeBoxIds.ToDictionary<string, string, long?>(keySelector: x => x, elementSelector: null);
            }
            return toReturn;
        }

        //private async Task<Dictionary<string, long?>> GetOnlineRecordId(List<string> chargeBoxIds)
        //{
        //    Dictionary<string, long?> toReturn = new();
        //    string sqlString = """
        //        SELECT Id
        //        FROM [dbo].[EVSEOnlineRecord] s
        //        WHERE ChargeBoxId = @ChargeBoxId
        //        ORDER BY s.OnlineTime DESC
        //        """;

        //    try
        //    {
        //        using var dbConn = await onlineLogDbConnectionFactory.CreateAsync();

        //        foreach(var chargeBoxId in chargeBoxIds)
        //        {
        //            var parms = new DynamicParameters();
        //            parms.Add("@ChargeBoxId", chargeBoxId, size: 36);
        //            var queryResult = await dbConn.QueryFirstOrDefaultAsync<long?>( sqlString, parms );
        //            toReturn.Add(chargeBoxId, queryResult);
        //        }
        //    }
        //    catch ( Exception e )
        //    {
        //        logger.LogError(e.ToString());
        //    }
        //    return toReturn;
        //}


        private bool IsOnlineNow(EVSECurrentStatus currentEVSE)
        {
            bool isOnline = false;
            int heartbeatInterval = 60;
            int.TryParse(currentEVSE.HeartbeatInterval, out heartbeatInterval);
            var checkTime = DateTime.UtcNow.AddSeconds(-(heartbeatInterval * 2));

            if (currentEVSE.HeartbeatUpdatedOn > checkTime)
            {
                isOnline = true;

            }

            return isOnline;

        }
    }
}
