﻿using EVCB_OCPP.TaskScheduler.Models;
using Newtonsoft.Json;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Transactions;
using Microsoft.Extensions.Logging;
using Microsoft.Extensions.DependencyInjection;

namespace EVCB_OCPP.TaskScheduler.Services
{
    public class CommonCustomerService : ICustomerService
    {
        private readonly ILogger<CommonCustomerService> logger;
        private readonly DatabaseService _dbService;// = new DatabaseService();
        private readonly OuterHttpClient httpClient;//= new OuterHttpClient();
        private Guid customerId = Guid.Empty;
        private string customerName = string.Empty;
        private string _partnerAPIRoot = string.Empty;
        private string _saltkey = string.Empty;
        private CancellationToken _ct;
        private ParallelOptions po = new ParallelOptions();
        private int ChargeRecordCallCounter = 0;

        public CommonCustomerService(
            DatabaseService databaseService,
            ILogger<CommonCustomerService> logger,
            ILogger<OuterHttpClient> clogger
            )
        {
            this._dbService = databaseService;
            this.logger = logger;// loggerProvider.CreateLogger(nameof(CommonCustomerService));
            this.httpClient = new OuterHttpClient(clogger);
        }

        public async Task SetCustomerId(Guid customerId)
        {
            this.customerId = customerId;

            //customerName = await _dbService.GetCustomerName(this.customerId);
            //_dbService.GetCustomerName(this.customerId);
            var connectionInfo = await _dbService.GetAPIConnectionInfo(customerId);
            customerName = connectionInfo.Name;
            _saltkey = connectionInfo.ApiKey;
            _partnerAPIRoot = connectionInfo.ApiUrl;
        }

        async public Task ReportStartTransaction(List<Models.Transaction> items = null)
        {
            //List<Models.Transaction> items = await _dbService.GetNeedReportSession(customerId, true, 1000);
            if (items == null)
            {
                items = await _dbService.GetNeedReportSession(customerId, true, 1000);
            }
            items = items.Where(x => x.CustomerId == customerId).ToList();

            Stopwatch watch = Stopwatch.StartNew();

            List<Task> groupTasks = new List<Task>();
            int completedCount = 0;
            int count = items.Count / 5 <= 100 ? items.Count : items.Count / 5;
            //var groupedItems = items.Chunk(50);

            while (completedCount < items.Count)
            {
                if (items.Count - completedCount < count)
                {
                    count = items.Count - completedCount;
                }

                var templst = items.Skip(completedCount).Take(count).ToList();
                if (templst.Count > 0)
                {
                    Task t = Task.Factory.StartNew(async () =>
                    {
                        await Assigned_StartTransactionCallbackTask(templst);
                    }, TaskCreationOptions.AttachedToParent);
                    groupTasks.Add(t);

                }
                completedCount += count;
            }
            while (ChargeRecordCallCounter != groupTasks.Count)
            {
                await Task.Delay(10);
            }

            watch.Stop();

            logger.LogDebug("ReportStartTransaction Task(" + items.Count() + ") : It takes  " + watch.ElapsedMilliseconds / 1000 + " Seconds");

        }

        async private Task Assigned_StartTransactionCallbackTask(List<Models.Transaction> reportlst)
        {
            await Task.Factory.StartNew(async () =>
            {
                //處理主機傳送的有指令
                try
                {
                    if (reportlst.Count > 0)
                    {

                        int completecounter = 0;
                        Dictionary<int, TransactionResponse> sendBack = new Dictionary<int, TransactionResponse>();

                        object responseLock = new object();
                        po.CancellationToken = _ct;
                        po.MaxDegreeOfParallelism = System.Environment.ProcessorCount;
                        Parallel.ForEach(reportlst, po, async (r) =>
                        {
                            var request = new
                            {
                                ChargeBoxId = r.ChargeBoxId,
                                ConnectorId = r.ConnectorId,
                                SessionId = r.Id,
                                MeterStart = r.MeterStart,
                                IdTag = r.StartIdTag,
                                StartTime = r.StartTime,
                                ReservationId = r.ReservationId
                            };

                            var response = await httpClient.Post(_partnerAPIRoot + "start_session", new Dictionary<string, string>()
                            {
                                { "PartnerId",customerId.ToString()}

                            }, JsonConvert.SerializeObject(request, DefaultSetting.JSONSERIALIZER_FORMAT), _saltkey);


                            lock (responseLock)
                            {
                                sendBack.Add(r.Id, new TransactionResponse()
                                {
                                    StartTransactionReportedOn = DateTime.UtcNow,
                                    ErrorMsg = response.Success ? null :
                                    (response.Exception == null ? response.Response : response.Exception.ToString())
                                });
                                Interlocked.Increment(ref completecounter);
                            }

                        });

                        while (completecounter != reportlst.Count)
                        {
                            await Task.Delay(10);
                        }

                        await _dbService.ReportStartTx(sendBack);

                    }
                }
                catch (Exception ex)
                {
                    logger.LogError("Assigned_StartTransactionCallbackTask Exception: " + ex.GetBaseException().ToString());
                }
                Interlocked.Increment(ref ChargeRecordCallCounter);
            
            }, TaskCreationOptions.AttachedToParent);


        }

        public async Task ReportStopTransaction(List<Models.Transaction> items = null)
        {
            if (items == null)
            {
                items = await _dbService.GetNeedReportSession(customerId, true, 1000);
            }
            items = items.Where(x => x.CustomerId == customerId).ToList();

            Stopwatch watch = new Stopwatch();
            watch.Start();

            List<Task> groupTasks = new List<Task>();
            int skipCount = 0;
            int count = items.Count / 5 <= 100 ? items.Count : items.Count / 5;
            while (skipCount < items.Count)
            {
                if (items.Count - skipCount < count)
                {
                    count = items.Count - skipCount;
                }
                var templst = items.Skip(skipCount).Take(count).ToList();
                if (templst.Count > 0)
                {
                    Task t = Task.Factory.StartNew(async () =>
                    {
                        await Assigned_StopTransactionCallbackTask(templst);
                    }, TaskCreationOptions.AttachedToParent);
                    groupTasks.Add(t);

                }
                skipCount += count;
            }
            while (ChargeRecordCallCounter != groupTasks.Count)
            {
                if (watch.ElapsedMilliseconds / 1000 > 300)
                {
                    logger.LogError(customerId + " ReportStopTransaction Lock "+ ChargeRecordCallCounter+" "+ groupTasks.Count);
                    break;
                }
                await Task.Delay(10);
            }

            watch.Stop();

            logger.LogDebug("ReportStopTransaction Task(" + items.Count() + ") : It takes  " + watch.ElapsedMilliseconds / 1000 + " Seconds");

        }

        async private Task Assigned_StopTransactionCallbackTask(List<Models.Transaction> reportlst)
        {
            await Task.Factory.StartNew(async () =>
            {
                //處理主機傳送的有指令
                try
                {
                    if (reportlst.Count > 0)
                    {

                        int completecounter = 0;
                        Dictionary<int, TransactionResponse> sendBack = new Dictionary<int, TransactionResponse>();

                        object responseLock = new object();
                        po.CancellationToken = _ct;
                        po.MaxDegreeOfParallelism = System.Environment.ProcessorCount;
                        Parallel.ForEach(reportlst, po, async (r) =>
                        {

                            if (!string.IsNullOrEmpty(r.Fee) && string.IsNullOrEmpty(r.Receipt))
                            {
                                //如果有計費但還沒結算 不回報給客戶                              
                                Interlocked.Increment(ref completecounter);
                            }
                            else
                            {
                                string CarNum = await _dbService.GetCustomerFromIdTag(r.StartIdTag);
                                var request = new
                                {
                                    ChargeBoxId = r.ChargeBoxId,
                                    ConnectorId = r.ConnectorId,
                                    SessionId = r.Id,
                                    MeterStart = r.MeterStart,
                                    MeterStop = r.MeterStop,
                                    IdTag = r.StartIdTag,
                                    StartTime = r.StartTime,
                                    StopTime = r.StopTime,
                                    StopReason = r.StopReasonId < 1 ? "Unknown" : (r.StopReasonId > 12 ? "Unknown" : ((Reason)r.StopReasonId).ToString()),
                                    ReservationId=r.ReservationId,
                                    Receipt = r.Receipt,
                                    TotalCost = r.Cost,
                                    Fee = r.Fee,
                                    CarNum = CarNum,

                                    StartSOC = r.StartSOC,
                                    StopSOC = r.StopSOC,
                                };

                                logger.LogDebug(JsonConvert.SerializeObject(request, DefaultSetting.JSONSERIALIZER_FORMAT));

                                var response = await httpClient.Post(_partnerAPIRoot + "completed_session", new Dictionary<string, string>()
                            {
                                { "PartnerId",customerId.ToString()}

                            }, JsonConvert.SerializeObject(request, DefaultSetting.JSONSERIALIZER_FORMAT), _saltkey);


                                lock (responseLock)
                                {
                                    sendBack.Add(r.Id, new TransactionResponse()
                                    {
                                        StopTransactionReportedOn = DateTime.UtcNow,
                                        ErrorMsg = response.Success ? null :
                                        (response.Exception == null ? response.Response : response.Exception.ToString())
                                    });
                                    Interlocked.Increment(ref completecounter);
                                   
                                }
                            }


                        });

                        while (completecounter != reportlst.Count)
                        {
                            logger.LogDebug(completecounter + "       " + reportlst.Count);
                            await Task.Delay(10000);
                        }

                        await _dbService.ReportStopTx(sendBack);







                    }
                }
                catch (Exception ex)
                {
                    logger.LogError("Assigned_StartTransactionCallbackTask Exception: " + ex.GetBaseException().ToString());
                }

                Interlocked.Increment(ref ChargeRecordCallCounter);
            }, TaskCreationOptions.AttachedToParent);


        }

        async public Task MonitorRemoteCommand()
        {
            Stopwatch watch = new Stopwatch();
            watch.Start();
            await _dbService.TurntoTimeoutMachineOperateCommands(10);
            await Task.Delay(10);
            watch.Stop();
            logger.LogDebug("ReportExecutionofRemoteCommand Task : It takes  " + watch.ElapsedMilliseconds / 1000 + " Seconds");

        }

        async public Task ReportExecutionofRemoteCommand()
        {
            var items = await _dbService.GetNeedReportExecution(customerId, 1000);

            Stopwatch watch = Stopwatch.StartNew();

            List<Task> groupTasks = new List<Task>();
            int skipCount = 0;
            int count = items.Count / 5 <= 100 ? items.Count : items.Count / 5;
            while (skipCount < items.Count)
            {
                if (items.Count - skipCount < count)
                {
                    count = items.Count - skipCount;
                }
                var templst = items.Skip(skipCount).Take(count).ToList();
                if (templst.Count > 0)
                {
                    Task t = Task.Factory.StartNew(async () =>
                    {
                        await Assigned_ReportExecutionofRemoteCommandTask(templst);
                    }, TaskCreationOptions.AttachedToParent);
                    groupTasks.Add(t);

                }
                skipCount += count;
            }
            while (ChargeRecordCallCounter != groupTasks.Count)
            {
                await Task.Delay(10);
            }

            watch.Stop();

            logger.LogDebug("ReportExecutionofRemoteCommand Task(" + items.Count() + ") : It takes  " + watch.ElapsedMilliseconds / 1000 + " Seconds");

        }

        async private Task Assigned_ReportExecutionofRemoteCommandTask(List<MachineOperateRecord> reportlst)
        {
            await Task.Factory.StartNew(async () =>
            {
                //處理主機傳送的有指令
                try
                {
                    if (reportlst.Count > 0)
                    {

                        int completecounter = 0;
                        Dictionary<int, BasicResponse> sendBack = new Dictionary<int, BasicResponse>();

                        object responseLock = new object();
                        po.CancellationToken = _ct;
                        po.MaxDegreeOfParallelism = System.Environment.ProcessorCount;
                        Parallel.ForEach(reportlst, po, async (r) =>
                        {
                            var request = new
                            {
                                ChargeBoxId = r.ChargeBoxId,
                                SerialNo = r.SerialNo,
                                CommandType = r.ActionConverttoCommandType(),
                                Result = r.GetExecution().Result,
                                Message = r.GetExecution().Detail,
                            };

                            var response = await httpClient.Post(_partnerAPIRoot + "commands/results", new Dictionary<string, string>()
                            {
                                { "PartnerId",customerId.ToString()}

                            }, JsonConvert.SerializeObject(request, DefaultSetting.JSONSERIALIZER_FORMAT), _saltkey);


                            lock (responseLock)
                            {
                                sendBack.Add(r.Id, new BasicResponse()
                                {
                                    ReportedOn = DateTime.UtcNow,
                                    ErrorMsg = response.Success ? null :
                                    (response.Exception == null ? response.Response : response.Exception.ToString())
                                });                              
                                Interlocked.Increment(ref completecounter);
                            }

                        });

                        while (completecounter != reportlst.Count)
                        {
                            await Task.Delay(10);
                        }

                        await _dbService.ReportExecution(sendBack);







                    }
                }
                catch (Exception ex)
                {
                    logger.LogDebug("Assigned_ReportExecutionofRemoteCommandTask Exception: " + ex.GetBaseException().ToString());
                }

                Interlocked.Increment(ref ChargeRecordCallCounter);
            }, TaskCreationOptions.AttachedToParent);


        }
    }
}
