﻿using Dapper;
using EVCB_OCPP.Domain.ConnectionFactory;
using EVCB_OCPP.TaskScheduler.Helper;
using EVCB_OCPP.TaskScheduler.Models;
using Microsoft.Extensions.Logging;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace EVCB_OCPP.TaskScheduler.Services;

public class OnlineLogDbService
{
    public OnlineLogDbService(
        ISqlConnectionFactory<OnlineLogDBContext> connectionFactory,
        ILogger<OnlineLogDbService> logger
        )
    {
        this.connectionFactory = connectionFactory;
        this.logger = logger;

        InitInsertOnlineLogHandler();
        InitUpdateOnlineLogHandler();
    }

    private void InitUpdateOnlineLogHandler()
    {
        if (updateOnlineLogHandler is not null)
        {
            throw new Exception($"{nameof(InitUpdateOnlineLogHandler)} should only called once");
        }

        updateOnlineLogHandler = new GroupHandler<EVSEOnlineRecord>(
            //BulkInsertWithBulkCopy,
            BundleUpdateWithDapper,
            //loggerFactory.CreateLogger("InsertMeterValueHandler")
            logger,
            workerCnt: 10
            );
    }

    private void InitInsertOnlineLogHandler()
    {
        if (insertOnlineLogHandler is not null)
        {
            throw new Exception($"{nameof(InitInsertOnlineLogHandler)} should only called once");
        }

        insertOnlineLogHandler = new GroupHandler<EVSEOnlineRecord>(
            //BulkInsertWithBulkCopy,
            BundleInsertWithDapper,
            //loggerFactory.CreateLogger("InsertMeterValueHandler")
            logger,
            workerCnt: 10
            );
    }

    private readonly ISqlConnectionFactory<OnlineLogDBContext> connectionFactory;
    private readonly ILogger<OnlineLogDbService> logger;

    private GroupHandler<EVSEOnlineRecord> insertOnlineLogHandler;
    private GroupHandler<EVSEOnlineRecord> updateOnlineLogHandler;

    internal Task UpdateOnlineLog(string chargeBoxId, DateTime offlineTime)
    {
        return updateOnlineLogHandler.HandleAsync(new EVSEOnlineRecord() { ChargeBoxId = chargeBoxId, OfflineTime = offlineTime });
        //return UpdateOnlineLogDapper(rowId, offlineTime);
    }

    internal Task InsertOnlineLog(string chargeBoxId, DateTime onlineTime)
    {
        return insertOnlineLogHandler.HandleAsync(new EVSEOnlineRecord() { ChargeBoxId = chargeBoxId, OnlineTime = onlineTime });
        //return InsertOnlineLogDapper(chargeBoxId, heartbeatUpdatedOn);
    }

    private async Task UpdateOnlineLogDapper(long rowId, DateTime offlineTime)
    {
        string sqlString = "UPDATE dbo.EVSEOnlineRecord SET OfflineTime=@OfflineTime WHERE Id=@Id";
        var parameters = new DynamicParameters();
        parameters.Add("@OfflineTime", offlineTime, System.Data.DbType.DateTime);
        parameters.Add("@Id", rowId, System.Data.DbType.Int64);

        using var dbConn = await connectionFactory.CreateAsync();
        await dbConn.ExecuteAsync(sqlString, parameters);
    }

    private async Task InsertOnlineLogDapper(string chargeBoxId, DateTime heartbeatUpdatedOn)
    {
        string sqlString = """
            INSERT INTO dbo.EVSEOnlineRecord ("ChargeBoxId","OnlineTime","OfflineTime")
            VALUES( @ChargeBoxId, @OnlineTime, @OfflineTime);
            """;
        var parameters = new DynamicParameters();
        parameters.Add("@ChargeBoxId", chargeBoxId, System.Data.DbType.String, System.Data.ParameterDirection.Input, 36);
        parameters.Add("@OnlineTime", heartbeatUpdatedOn, System.Data.DbType.DateTime);
        parameters.Add("@OfflineTime", DefaultSetting.DefaultNullTime, System.Data.DbType.DateTime);

        using var dbConn = await connectionFactory.CreateAsync();
        await dbConn.ExecuteAsync(sqlString, parameters);
    }

    private async Task BundleUpdateWithDapper(BundleHandlerData<EVSEOnlineRecord> bundleHandlerData)
    {
        string sqlString = """
            UPDATE dbo.EVSEOnlineRecord 
            SET OfflineTime=@OfflineTime 
            WHERE Id = (
                SELECT TOP(1) Id
                FROM dbo.EVSEOnlineRecord
                WHERE ChargeBoxId = @ChargeBoxId
                ORDER BY OnlineTime DESC
            )
            """;

        using var dbConn = await connectionFactory.CreateAsync();
        foreach (var record in bundleHandlerData.Datas)
        {
            var parameters = new DynamicParameters();
            parameters.Add("@ChargeBoxId", record.ChargeBoxId, System.Data.DbType.String, System.Data.ParameterDirection.Input, 36);
            parameters.Add("@OfflineTime", record.OfflineTime, System.Data.DbType.DateTime);
            //parameters.Add("@Id", record.Id, System.Data.DbType.Int64);
            await dbConn.ExecuteAsync(sqlString, parameters);
            bundleHandlerData.AddCompletedData(record);
        }
    }

    private async Task BundleInsertWithDapper(BundleHandlerData<EVSEOnlineRecord> bundleHandlerData)
    {
        string sqlString = """
            INSERT INTO dbo.EVSEOnlineRecord ("ChargeBoxId","OnlineTime","OfflineTime")
            VALUES( @ChargeBoxId, @OnlineTime, @OfflineTime);
            """;

        using var dbConn = await connectionFactory.CreateAsync();
        foreach (var record in bundleHandlerData.Datas)
        {
            var parameters = new DynamicParameters();
            parameters.Add("@ChargeBoxId", record.ChargeBoxId, System.Data.DbType.String, System.Data.ParameterDirection.Input, 36);
            parameters.Add("@OnlineTime", record.OnlineTime, System.Data.DbType.DateTime);
            parameters.Add("@OfflineTime", DefaultSetting.DefaultNullTime, System.Data.DbType.DateTime);
            await dbConn.ExecuteAsync(sqlString, parameters);
            bundleHandlerData.AddCompletedData(record);
        }
    }
}
